//-------------------------------------------------------------
// Hidden Conditional Random Field Library - GradientSHDCRF
// Component
//
//	June, 2014
//	Michael Soukup

#include "gradient.h"
#include <exception>

GradientSHDCRF::GradientSHDCRF(InferenceEngine* infEngine,
							 FeatureGenerator* featureGen):
Gradient(infEngine, featureGen)
{
}

double GradientSHDCRF::computeGradient(dVector& vecGradient, Model* m,
									DataSequence* X)
{
	//compute beliefs
	Beliefs bel;
	Beliefs belMasked;
	pInfEngine->computeBeliefs(bel, pFeatureGen, X, m, true, -1, false);
	pInfEngine->computeBeliefs(belMasked, pFeatureGen, X, m, true, -1, true);

	//estimate p(y_t|h_t)
	dMatrix ybel;
	pFeatureGen->computeLabelBeliefs(ybel, X, m);

	//Check the size of vecGradient
	int nbFeatures = pFeatureGen->getNumberOfFeatures();
	if(vecGradient.getLength() != nbFeatures)
		vecGradient.create(nbFeatures);

#if !defined(_VEC_FEATURES) && !defined(_OPENMP)
	featureVector* vecFeatures;
#endif
#if defined(_OPENMP)
	int ThreadID = omp_get_thread_num();
	if (ThreadID >= nbThreadsMP)
		ThreadID = 0;
#else
	int ThreadID = 0;
#endif
	//Loop over nodes to compute features and update the gradient
	double phi = 0.0;
	double Zx = 0.0;
	for(int i = 0; i < X->length(); i++)
	{
		int s = X->getStateLabels(i);
		for (int yp = 0; yp < m->getNumberOfStateLabels(); yp++) {
			for (int hp = 0; hp < m->getNumberOfStates(); hp++) {
				if (s == yp) {
					phi += belMasked.belStates[i][hp] * ybel.getValue(yp, hp);
				}
				Zx += bel.belStates[i][hp] * ybel.getValue(yp, hp);
			}
		}

		//Get nodes features
#if defined(_VEC_FEATURES) || defined(_OPENMP)
		pFeatureGen->getFeatures(vecFeaturesMP[ThreadID], X,m,i,-1);
		// Loop over features
		feature* pFeature = vecFeaturesMP[ThreadID].getPtr();
		for(int j = 0; j < vecFeaturesMP[ThreadID].size(); j++, pFeature++)
#else
		vecFeatures = pFeatureGen->getFeatures(X,m,i,-1);
		// Loop over features
		feature* pFeature = vecFeatures->getPtr();
		for(int j = 0; j < vecFeatures->size(); j++, pFeature++)
#endif
		{
			//p(y_i=s|x)*f_k(i,s,x)
			vecGradient[pFeature->globalId] += belMasked.belStates[i][pFeature->nodeState]*pFeature->value;
			vecGradient[pFeature->globalId] -= bel.belStates[i][pFeature->nodeState]*pFeature->value;
		}

		// Read state label
#if defined(_VEC_FEATURES) || defined(_OPENMP)
		pFeatureGen->getFeatures(vecFeaturesMP[ThreadID], X,m,i,i+X->length());
		// Loop over features
		pFeature = vecFeaturesMP[ThreadID].getPtr();
		for(int j = 0; j < vecFeaturesMP[ThreadID].size(); j++, pFeature++)
#else
		vecFeatures = pFeatureGen->getFeatures(X,m,i,i+X->length());
		// Loop over features
		pFeature = vecFeatures->getPtr();
		for(int j = 0; j < vecFeatures->size(); j++, pFeature++)
#endif
		{

			// Update gradient only on equal label
			// This makes p(h_t, y_t) = 1
			if (pFeature->prevNodeState == s) {
				//p(y_i=s|x)*g_k(i,s,y)
				vecGradient[pFeature->globalId] += belMasked.belStates[i][pFeature->nodeState]*pFeature->value;
				//vecGradient[pFeature->globalId] -= bel.belStates[i][pFeature->nodeState]*pFeature->value;
				//vecGradient[pFeature->globalId] -= ybel.getValue(pFeature->prevNodeState, pFeature->nodeState)*bel.belStates[i][pFeature->nodeState]*pFeature->value*bel.belStates[i][pFeature->nodeState];
				//vecGradient[pFeature->globalId] -= ybel.getValue(pFeature->prevNodeState, pFeature->nodeState)*bel.belStates[i][pFeature->nodeState]*pFeature->value;
			}
			vecGradient[pFeature->globalId] -= ybel.getValue(pFeature->prevNodeState, pFeature->nodeState)*bel.belStates[i][pFeature->nodeState]*pFeature->value;
		}
	}


	for(int i = 0; i < X->length()-1; i++) // Loop over all rows (the previous node index)
	{
		//Get nodes features
#if defined(_VEC_FEATURES) || defined(_OPENMP)
		pFeatureGen->getFeatures(vecFeaturesMP[ThreadID], X,m,i+1,i);
		// Loop over features
		feature* pFeature = vecFeaturesMP[ThreadID].getPtr();
		for(int j = 0; j < vecFeaturesMP[ThreadID].size(); j++, pFeature++)
#else
		vecFeatures = pFeatureGen->getFeatures(X,m,i+1,i);
		// Loop over features
		feature* pFeature = vecFeatures->getPtr();
		for(int j = 0; j < vecFeatures->size(); j++, pFeature++)
#endif
		{
			//p(y_i=s1,y_j=s2|x)*f_k(i,j,s1,s2,x) is subtracted from the gradient 
			vecGradient[pFeature->globalId] += belMasked.belEdges[i](pFeature->prevNodeState,pFeature->nodeState)*pFeature->value;
			vecGradient[pFeature->globalId] -= bel.belEdges[i](pFeature->prevNodeState,pFeature->nodeState)*pFeature->value;
		}
	}

	//return log(Zx) +  bel.partition - belMasked.partition;

	//Return -log instead of log() [Moved to Gradient::ComputeGradient by LP]
	//vecGradient.negate();

	//double partition = pInfEngine->computePartition(pFeatureGen, X, m, -1, true);
	//std::cout << "Phi: " << log(phi) << std::endl;
	//std::cout << "Partition: " << log(Zx) << std::endl;

	//return log(Zx) - log(phi);

	//double funcval = pFeatureGen->computeLogObjective(X, m);
	//return log(Zx) - log(phi) + partition - bel.partition; // + bel.partition - belMasked.partition
	//return -funcval;
	return log(Zx) - log(phi) + bel.partition;
}

// This is done. Maybe double-check on entropy calculations.
double GradientSHDCRF::computeGradient(dVector& vecGradient, Model* m, DataSet* X)
{

	//Check the size of vecGradient
	int nbFeatures = pFeatureGen->getNumberOfFeatures();
	double ans = 0.0;
	int TID = 0;
	if(vecGradient.getLength() != nbFeatures)
		vecGradient.create(nbFeatures);
	else
		vecGradient.set(0);
	// Initialize the buffers (vecFeaturesMP) for each thread
#ifdef _OPENMP
	setMaxNumberThreads(omp_get_max_threads());
	pInfEngine->setMaxNumberThreads(omp_get_max_threads());
	pFeatureGen->setMaxNumberThreads(omp_get_max_threads());
#endif
	for(int t=0;t<nbThreadsMP;t++)
	{
		if(localGrads[t].getLength() != nbFeatures)
			localGrads[t].resize(1,nbFeatures,0);
		else
			localGrads[t].set(0);
	}

////////////////////////////////////////////////////////////
// Start of parallel Region

	// Some weird stuff in gcc 4.1, with openmp 2.5 support
#if ((_OPENMP == 200505) && __GNUG__)
#pragma omp parallel \
	shared(X, m, ans, nbFeatures, std::cout)	\
	private(TID) \
	default(none)
#else
#pragma omp parallel \
	shared(vecGradient, X, m, ans, nbFeatures, std::cout)	\
	private(TID) \
	default(none)
#endif
	{
#ifdef _OPENMP 
		TID = omp_get_thread_num();
#endif
		// Create a temporary gradient
		double localSum = 0;

#ifdef WITH_SEQUENCE_WEIGHTS
		dVector tmpVecGradient(nbFeatures);
#endif

#pragma omp for
		// we can use unsigned if we have openmp 3.0 support (_OPENMP>=200805).
#ifdef _OPENMP 
    #if _OPENMP >= 200805
		for(unsigned int i = 0; i< X->size(); i++){
    #else
	    for(int i = 0; i< X->size(); i++){
    #endif
#else
		for(unsigned int i = 0; i< X->size(); i++){
#endif
			if (m->getDebugLevel() >=3){
#pragma omp critical(output)
				std::cout << "Thread "<<TID<<" computes gradient for sequence " 
						  << i <<" out of " << (int)X->size() 
						  << " (Size: " <<  X->at(i)->length() << ")" << std::endl;
			}
			DataSequence* x = X->at(i);
#ifdef WITH_SEQUENCE_WEIGHTS
			tmpVecGradient.set(0);
			localSum += computeGradient(tmpVecGradient, m, x) * x->getWeightSequence();
			if(x->getWeightSequence() != 1.0)
				tmpVecGradient.multiply(x->getWeightSequence());
			localGrads[TID].add(tmpVecGradient);
#else
			localSum += computeGradient(localGrads[TID], m, x);// * x->getWeightSequence();
#endif
		}
#pragma omp critical (reduce_sum)
		// We now put togheter the sums
		{
			if( m->getDebugLevel() >= 2){
				std::cout<<"Thread "<<TID<<" update sums"<<std::endl;
			}
			ans += localSum;
			vecGradient.add(localGrads[TID]);
		}
	} 
	
// End of parallel Region
////////////////////////////////////////////////////////////


	// Add conditional entropy terms
	dVector* beta = m->getWeights();
	double H_lambda = 0.0;
	dMatrix ybel;
	dVector dH_sum(m->getNumberOfStates());
	
	pFeatureGen->computeLabelBeliefs(ybel, X->at(0), m); // Computes p(y_i|h_j). Seqmul?? TODO
	dH_sum.set(0);

	// Loop over latent label features
#if defined(_VEC_FEATURES) || defined(_OPENMP)
	pFeatureGen->getFeatures(vecFeaturesMP[ThreadID], X->at(0), m, 0, X->at(0)->length(), -1);
	feature* pFeature = vecFeaturesMP[ThreadID].getPtr();
	for (int j = 0; j < vecFeaturesMP[ThreadID].size() ; j++, pFeature++)
#else
	featureVector* vecFeature = pFeatureGen->getFeatures(X->at(0), m, 0, X->at(0)->length() , -1);
	feature* pFeature = vecFeature->getPtr();
	for (int j = 0; j < vecFeature->size() ; j++, pFeature++)
#endif
	{
		int id = pFeature->globalId;
		int hp = pFeature->nodeState;
		int yp = pFeature->prevNodeState;
		double p_yh = ybel.getValue(yp, hp);
		H_lambda -= p_yh * log(p_yh);	
		dH_sum.setValue(hp, dH_sum.getValue(hp) + p_yh*(*beta)[id]);
	}

	// Loop over again to compute gradient contributions
#if defined(_VEC_FEATURES) || defined(_OPENMP)
	pFeatureGen->getFeatures(vecFeaturesMP[ThreadID], X->at(0), m, 0, X->at(0)->length(), -1);
	pFeature = vecFeaturesMP[ThreadID].getPtr();
	for (int j = 0; j < vecFeaturesMP[ThreadID].size() ; j++, pFeature++)
#else
	vecFeature = pFeatureGen->getFeatures(X->at(0), m, 0, X->at(0)->length(), -1);
	pFeature = vecFeature->getPtr();
	for (int j = 0; j < vecFeature->size() ; j++, pFeature++)
#endif
	{
		int id = pFeature->globalId;
		int hp = pFeature->nodeState;
		int yp = pFeature->prevNodeState;
		double p_yh = ybel.getValue(yp, hp);
		vecGradient[id] += p_yh * ((*beta)[id] - dH_sum.getValue(hp));
	}

	// Finally, add the entropy-based term to function value
	ans += m->getSparsityFactor()*H_lambda;
	std::cout<<"entropy "<< H_lambda << std::endl;

	// because we are minimizing -LogP
	vecGradient.negate();


	// Add the regularization term
	double sigmaL2Square = m->getRegL2Sigma()*m->getRegL2Sigma();
	if(sigmaL2Square != 0.0f) {
		double weightNorm = m->getWeights()->l2Norm(false);
		double normsquared = m->getWeights()->l2Norm(true);
		std::cout << "weightnorm " << normsquared << std::endl
				  << "L2reg " << (weightNorm / (2.0*sigmaL2Square)) << std::endl;
		ans += weightNorm / (2.0*sigmaL2Square);	

		std::cout<<"weights ";
		std::cout.precision(4);
		for(int f = 0; f < nbFeatures; f++) {
			vecGradient[f] += (*beta)[f]/sigmaL2Square;
			std::cout <<  (*beta)[f] <<  " ";
		}

		std::cout<<"\ngradients ";
		for(int f = 0; f < nbFeatures; f++) {
			std::cout <<  vecGradient[f] <<  " ";
		}		
	}


	// std::cout << "\np(y_i | h_j): " << std::endl;
	// for (int hp = 0; hp < m->getNumberOfStates(); hp++) {
	// 	std::cout << "hp = " << hp << ": ";
	// 	for (int yp = 0; yp < m->getNumberOfStateLabels(); yp++) {
	// 		std::cout << ybel.getValue(yp, hp) << "  ";
	// 	}
	// 	std::cout << std::endl;
	// }
	std::cout << "\nf " << ans << "\n" << std::endl;	

	return ans;
}
