//-------------------------------------------------------------
// Hidden Conditional Random Field Library - EvaluatorSHDCRF
// Component
//
//	June, 2014
//	Michael Soukup

#include "evaluator.h"
#include <assert.h>
using namespace std;

/////////////////////////////////////////////////////////////////////
// Evaluator SHDCRF Class
/////////////////////////////////////////////////////////////////////

// *
// Constructor and Destructor
// *

EvaluatorSHDCRF::EvaluatorSHDCRF() : Evaluator()
{
}

EvaluatorSHDCRF::EvaluatorSHDCRF(InferenceEngine* infEngine, FeatureGenerator* featureGen)
: Evaluator(infEngine, featureGen)
{
}

EvaluatorSHDCRF::~EvaluatorSHDCRF()
{

}
// *
// Public Methods
// *

double EvaluatorSHDCRF::computeError(DataSequence* X, Model* m)
{
	if(!pInfEngine || !pFeatureGen){
		throw HcrfBadPointer("In EvaluatorSHDCRF::computeError");
	}

	double partitionMasked = pInfEngine->computePartition(pFeatureGen, X, m, -1, true);
	double partition = pInfEngine->computePartition(pFeatureGen, X, m, -1, false);
	// return log(Z(h|x)) - log(Z(h*|x)) 
	return partition - partitionMasked;
}

double EvaluatorSHDCRF::computeError(DataSet* X, Model* m)
{
	if(!pInfEngine || !pFeatureGen){
		throw HcrfBadPointer("In EvaluatorSHDCRF::computeError(DataSet*, model*)");
	}

	double error = 0;
	int TID = 0;
	int NumIters = (int)X->size();

	// Initialize the buffers (vecFeaturesMP) for each thread
#ifdef _OPENMP
	pInfEngine->setMaxNumberThreads(omp_get_max_threads());
	pFeatureGen->setMaxNumberThreads(omp_get_max_threads());
#endif

#pragma omp parallel shared(std::cout, X, m, NumIters, error)	\
	private(TID) \
	default(none)
	{
#ifdef _OPENMP
		TID = omp_get_thread_num();
#else
		TID = 0;
#endif
#pragma omp for reduction(+:error)
		for(int i = 0; i<NumIters; i++)
		{
			if (m->getDebugLevel()>=4)
			{
#pragma omp critical(output)
				cout << "Thread "<<TID<<" Computing error for sequence " << i << " out of " 
					 << (int)X->size() << " (Size: " << X->at(i)->length() 
					 << ")" << endl;
			}
			error += this->computeError(X->at(i), m) * X->at(i)->getWeightSequence();
		}
	}
	//error = error/X->size(); // Make the influence of regularization factor to be independent of the size of training data  
	if(m->getRegL2Sigma() != 0.0f)
	{
		double weightNorm = m->getWeights()->l2Norm(false);
		error += weightNorm / (2.0*m->getRegL2Sigma()*m->getRegL2Sigma());
	}

	// TODO add entropy-based term 
	return error;
}


void EvaluatorSHDCRF::computeStateLabels(DataSequence* X, Model* m, iVector* vecStateLabels, dMatrix * probabilities)
/*
This function is used to compute the probability of each nodes in the
datasequence given the model. It return a vector of labels and the matrix
of probability
*/
{
	if(!pInfEngine || !pFeatureGen){
		throw HcrfBadPointer("EvaluatorSHDCRF::computeStateLabels");
	}

	int nbNodes = X->length();
	int nbLabels = m->getNumberOfStateLabels();
	int nbStates = m->getNumberOfStates();
	
	Beliefs bel;
	dMatrix ybel;
	pInfEngine->computeBeliefs(bel, pFeatureGen, X, m, false, -1, false);
	pFeatureGen->computeLabelBeliefs(ybel, X, m); // estimates p(y_i|h_j)

	vecStateLabels->create(nbNodes);
	if(probabilities)
		probabilities->create(nbNodes,nbLabels);

	dVector sumBeliefsPerLabel(nbLabels);
	for (int hp = 0; hp<nbStates; hp++) {
		for (int yp = 0; yp<nbLabels; yp++) {
			std::cout << ybel.getValue(yp, hp) << "  ";
		}
		std::cout << endl;
	}


	for(int n = 0; n<nbNodes; n++) {
		// Sum beliefs
		sumBeliefsPerLabel.set(0);
		for (int s = 0; s<nbStates; s++) {
			for (int y = 0; y<nbLabels; y++) {
				// p(y|h)*p(h|x)
				sumBeliefsPerLabel[y] += ybel.getValue(y, s) * bel.belStates[n][s];
			}
		}

		// find max value
		vecStateLabels->setValue(n, 0);
		double MaxBel = sumBeliefsPerLabel[0];
		if(probabilities)
			probabilities->setValue(0,n,sumBeliefsPerLabel[0]);
		for (int cur_label = 1; cur_label < nbLabels; cur_label++) {
			if(probabilities) {
				probabilities->setValue(cur_label, n, sumBeliefsPerLabel[cur_label]);
			}
			if(MaxBel < sumBeliefsPerLabel[cur_label]) {
				vecStateLabels->setValue(n, cur_label);
				MaxBel = sumBeliefsPerLabel[cur_label];
			}
		}
	}
}
