"""Michael Soukup 2014

BCIEEG handles EEG data in use with a BCI.
Designed for real-time usage. (Off-line analyses can easily be simulated though)

- Selective channels
- Feature extraction
- Data buffer
- ...

"""

import pyeeg
from collections import deque


class ChannelError(Exception):
	pass


class BCIEEG(object):

	# Labels, and thus location on the scalp, must be known for the module so that it can do ErrP detection and draw the right features.
	chan_labels = (
		'A3',
		'A4'
		#etc
	)

	freqbands = [0.5,4,7,12,30] # delta, theta, alpha and beta respectively

	features = (
		"bandpower",
		"bandpower_rel"
	)

	def __init__(self, chanlist, fs, windowsize=1):
		# chanlist: Ordered list of channel labels that correspond to incoming data
		# fs: Sample frequency
		# windowsize: Size of window (s) for feature extraction

		self.N_chan = len(chanlist)
		# chanmap: maps channel labels to the right index (for setting mask more intuitively)
		self.chanmap = {}
		for i,c in enumerate(chanlist):
			if c not in BCIEEG.chan_labels:
				raise ChannelError('Channel "%s" not know' % c)
			self.chanmap[c] = i

		# check channels for motor imagery and errp detection, then set masks.
		self.errp_detection = self.chanmap.has_key('errp det channels')
		self.fmask = range(self.N_chan) # TODO set according to mental strategy and/or errp detection

		# TODO make the set of features selectable. Use a initial default set
		# self.features = (..., ...)

		self.fs = fs
		self.dt = 1.0/fs
		self.windowsize = windowsize
		self.N_window = int(windowsize*fs)

		# Init data to zeros. TODO check minimum samples for ErrP detection
		self.data = deque([[0 for i in range(self.N_chan)] for j in range(self.N_window)], maxlen=self.N_window)


	def set_channel_mask(self, chans):
		# To set manually or internally
		# chans: List of channel labels for feature extraction. Must be international 10-20 system.
		self.fmask = [self.chanmap[c] for c in chans]


	def add_data(self, data):
		# data: [
		# 	[chan_1, chan_2, ..., chan_n], t=0
		# 	[chan_1, chan_2, ..., chan_n], t=dt
		# 	[chan_1, chan_2, ..., chan_n], t=2*dt
		# 	...
		# 	[chan_1, chan_2, ..., chan_n], t=N*dt
		# ]
		for segment in data:
			self.data.appendleft(segment)


	def get_features(self):
		fvec = []
		for i in self.fmask:
			chandata = [j[i] for j in self.data] # reverse?
			chandata.reverse()
			_, bp_rel = pyeeg.bin_power(chandata, BCIEEG.freqbands, self.fs)
			for b in bp_rel:
				fvec.append(b)
		return fvec


